// General blurring routine using a kernel in kernelFluxArray on energies
// kernelEnergyArray operating on fluxArray on energies energyArray.
// This is used by rdblur, kdblur, and kdblur2

#include <xsTypes.h>
#include <XSUtil/Numerics/BinarySearch.h>
#include <iostream>
#include <limits>

void blurring(const RealArray& kernelEnergyArray, const RealArray& kernelFluxArray,
	      const Real kernelLineEnergy, const RealArray& energyArray, RealArray& fluxArray,
	      const bool Reset)
{
  //  static bool first(true);

  // fract is the fraction of the line shape integral used in the convolution
  const Real fract(0.99999);

  size_t kne(kernelFluxArray.size());
  size_t halfkne = kne/2;

  // normalize line to an integral of 1
  RealArray normKernelFluxArray(kne);
  Real sum = kernelFluxArray.sum();
  if ( sum == 0.0 ) return;
  normKernelFluxArray = kernelFluxArray/sum;

  // find range which encompasses fract of the area under
  // the line symmetrically from the center

  sum = 0.0;
  size_t ilow = kne/2-1;
  size_t ihigh = kne/2;
  const size_t iLimit = std::numeric_limits<std::size_t>::max();
  while ( sum <= fract && ilow >= 0 && ihigh < kne ) {
    if ( ilow != iLimit ) {
      sum += normKernelFluxArray[ilow];
      ilow--;
    }
    if ( ihigh <= kne ) {
      sum += normKernelFluxArray[ihigh];
      ihigh++;
    }
  }

  if ( ilow == iLimit ) ilow = 0;
  if ( ihigh > kne-1 ) ihigh = kne-1;

  Real eratlow = kernelEnergyArray[ilow]/kernelLineEnergy;
  Real erathigh = kernelEnergyArray[ihigh]/kernelLineEnergy;

  // Initialize the convolution output

  size_t ne(fluxArray.size());
  RealArray convolvedFluxArray(ne);
  convolvedFluxArray = 0.0;

  // Loop over energy ranges for the current dataset

  size_t jlow = 0;
  size_t jhigh = 0;

  for (size_t ie=0; ie<ne; ie++) {

    Real energy=0.5*(energyArray[ie]+energyArray[ie+1]);

    // find first and last bins required in the convolution

    Real estart = energy * eratlow;
    Real estop = energy * erathigh;

    while ( energyArray[jlow] < estart && jlow < ne ) jlow++;

    if ( jlow > 0 ) jlow--;
    while ( energyArray[jhigh] < estop && jhigh < ne-1 ) jhigh++;

    for (size_t je=jlow; je<=jhigh; je++) {

      Real f1 = energyArray[je]/energy*halfkne*(1.0+1.0/halfkne);
      Real f2 = energyArray[je+1]/energy*halfkne*(1.0+1.0/halfkne);

      int if1 = (int)f1;
      Real r1 = f1 - if1;
      int if2 = (int)f2;
      Real r2 = f2 - if2;
      sum = 0.0;

      for (int i=if1+1; i<=if2-1; i++) {
	if ( i >= 0 && i < kne ) sum += normKernelFluxArray[i];
      }
             
      if ( if1 >= 0 && if1 < kne ) sum += (1.0-r1)*normKernelFluxArray[if1];
             
      if ( if2 >= 0 && if2 < kne ) sum += (r2)*normKernelFluxArray[if2];
             
      if ( if1 == if2 ) {
	if ( if1 >= 0 && if1 < kne ) sum -= normKernelFluxArray[if1];
      }

      convolvedFluxArray[je] += fluxArray[ie]*sum;
    }
  }

  fluxArray = convolvedFluxArray;
  
  return;
}

// Version of the routine that can be called from C or Fortran

void doblurring(const float* kear, const int kne, const float* kflux,
		const float klineE, const float* ear, const int ne, float* flux,
		const bool reset)
{
  RealArray kernelEnergyArray(kne+1);
  RealArray kernelFluxArray(kne);
  RealArray energyArray(ne+1);
  RealArray fluxArray(ne);
  
  Real kernelLineEnergy((Real)klineE);

  if ( reset ) {
    for (int i=0; i<kne+1; i++) kernelEnergyArray[i] = kear[i];
    for (int i=0; i<kne; i++) kernelFluxArray[i] = kflux[i];
  }
  for (int i=0; i<ne; i++) fluxArray[i] = flux[i];
  for (int i=0; i<ne+1; i++) energyArray[i] = ear[i];

  blurring(kernelEnergyArray, kernelFluxArray, kernelLineEnergy, energyArray,
	   fluxArray, reset);

  for (int i=0; i<ne; i++) flux[i] = fluxArray[i];
  
  return;
}

#include <cfortran.h>
void doblurring(const float* kear, const int kne, const float* kflux,
		const float klineE, const float* ear, const int ne, float* flux, const bool reset);
FCALLSCSUB8(doblurring,DOBLURRING,doblurring,FLOATV,INT,FLOATV,FLOAT,FLOATV,INT,FLOATV,LOGICAL)
