!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!  The RELCONV MODEL (written by Thomas Dauser)
!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!  If you use this model for a publication, please cite our paper
!  "Dauser, Thomas; Wilms, J.; Reynolds, C. S.; Brenneman, L. W.,
!  2010, MNRAS, 409, 1534-1540"
!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!  This code, used together with relbase.f90, can be used to calculate
!  the relativistic smearing and is desinged to be used within 
!  a X-ray data analysis program like XSPEC or ISIS.
! 
!
!  See www.sternwarte.uni-erlangen.de/research/relline/ or 
!  "relbase.f90" for information on the RELLINE model
!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
MODULE par_conv
! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
! defines all parameters used in variuos routines

  IMPLICIT NONE
  SAVE
  ! ******** parameters for the convolution ***************** !!!
  integer :: halfne
  real, allocatable :: flu(:)
  real, allocatable :: cached_params(:),cphotar(:),cached_input(:)
  real, allocatable :: cached_ear(:)
  integer :: cached_ne
  logical :: first_cached = .true.
  ! *************************************************************** !!!!

END MODULE par_conv


! ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SUBROUTINE tdrelconv (ear, ne, param, ifl, photar, photer)
! ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  
  use par
  implicit none
  ! parameters required by XSPEC
  integer,parameter :: num_param = 8
  INTEGER :: ne, ifl
  REAL :: ear(0:ne), param(num_param)
  REAL :: photar(ne), photer(ne)
  ! further information required

  ! silly line to suppress a compiler warning. note that model_type is
  ! immediately overwritten
  model_type = ifl
  
  ! not lamp post geometry; normal powerlaw emissivity
  model_type = 1

  call relbase_conv(ear, ne, param, photar,num_param)
  
END SUBROUTINE tdrelconv ! +++++++++++++++++++++++++++++++++++++++


! ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SUBROUTINE tdrelconvlp (ear, ne, param, ifl, photar, photer)
! ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  
  use par
  implicit none
  ! parameters required by XSPEC
  integer,parameter :: num_param = 7
  INTEGER :: ne, ifl
  REAL :: ear(0:ne), param(num_param)
  REAL :: photar(ne), photer(ne)

  ! silly line to suppress a compiler warning. note that model_type is
  ! immediately overwritten
  model_type = ifl
  
  ! lamp post geometry
  model_type = 2

  call relbase_conv(ear, ne, param, photar,num_param)
  
END SUBROUTINE tdrelconvlp ! +++++++++++++++++++++++++++++++++++++


! ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SUBROUTINE tdrelconvlpext (ear, ne, param, ifl, photar, photer)
! ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  
  use par
  implicit none
  ! parameters required by XSPEC
  integer,parameter :: num_param = 10
  INTEGER :: ne, ifl
  REAL :: ear(0:ne), param(num_param)
  REAL :: photar(ne), photer(ne)

  ! silly line to suppress a compiler warning. note that model_type is
  ! immediately overwritten
  model_type = ifl
  
  ! extended
  model_type = 3

  call relbase_conv(ear, ne, param, photar,num_param)
  
END SUBROUTINE tdrelconvlpext ! +++++++++++++++++++++++++++++++++++++


! ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SUBROUTINE relbase_conv (ear, ne, param, photar,num_param)
! ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  
  USE par
  USE par_conv
  IMPLICIT NONE
  INTEGER :: ne
  integer :: num_param,num_param_relline
  REAL :: ear(0:ne), param(num_param)
  REAL :: photar(ne)
  REAL :: param_relline(num_param+2)
  integer :: i
  logical :: is_cached,same_flux

  ! convert the parameter for the RELLINE routine
  num_param_relline = num_param+2
  
  ! == try caching  == !
  ! caching done completely by relconv
  no_caching_relline = .true. 
  call check_caching(param,num_param,photar,ear,ne,is_cached,same_flux)

  ! if we can use the cached convolution grid 
  if ( is_cached ) then
     ! if the input flux is also the same then just copy the cached result
     ! and return otherwise call doblurring with the reset argument set to
     ! false. note that in this second case we don't need to bother calling
     ! allocating memory or calling relline
     if ( same_flux ) then
        do i=1,ne
           photar(i) = cphotar(i)
        end do
     else
        call doblurring(ear, ne, flu, param_relline(1), ear, ne, photar, .false.)
        ! cache the result
        do i=1,ne
           cphotar(i) = photar(i)
        enddo
     endif
     return
  endif
  ! ================== !

  ! * "model is not cached", i.e., we do not care if the input flux is the
  !   same
  ! * "if the model is already cached", but the flux is not the same
  !   we can use the variable grid to do the multiplication faster
  if (.not.is_cached) then 
     call init_par_conv(param, param_relline,num_param,num_param_relline,model_type)

     ! allocated space for the array to calculate the convolution function
     allocate(flu(ne))
     
     ! set the line energy for the relline model to the middle of the
     ! energy range. the exact energy doesn't really matter it just needs
     ! to be such that the non-zero part of the relline model is output
     param_relline(1) = ear(ne/2)
     
     ! call the RELLINE model
     ! CAVEAT: number of parameters in the relline model
     !         is two more than for the convolution model!
     call relbase(ear,ne,param_relline,flu,num_param_relline)

     ! do the convolution using a routine in blurring.cxx
     call doblurring(ear, ne, flu, param_relline(1), ear, ne, photar, .true.)

     ! tidy up memory
     deallocate(flu)
     
  end if
  
  ! cache the result
  do i=1,ne
     cphotar(i) = photar(i)
  enddo

     
END SUBROUTINE relbase_conv !+++++++++++++++++++++++++++++++++++++++++++




!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SUBROUTINE INIT_PAR_CONV(param, param_relline,num_param,num_param_relline,model_type)
!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!  this routine initializes the line parameters

  !---------------------------------
  IMPLICIT NONE
  integer :: num_param,num_param_relline
  real :: param(num_param)
  real :: param_relline(num_param_relline)
  double precision  :: emis1,emis2,rbreak,gamma
  double precision  :: a,theta0grad,rmin,rmax,h
  double precision  :: hbase,htop,vbase,v100
  integer :: limb_law,model_type
  !---------------------------------

  param_relline(1) = 1.D0  ! LineE

  if (model_type.eq.1) then
     ! READ Parameter Values
     emis1=param(1)
     emis2=param(2)
     rbreak=param(3)
     a=param(4)
     theta0grad=param(5)
     rmin=param(6)
     rmax=param(7)
     limb_law=int(param(8))
     
     
     ! WRITE Parameter values (for RELBASE routine)
     param_relline(2) = real(emis1)
     param_relline(3) = real(emis2)
     param_relline(4) = real(rbreak)
     param_relline(5) = real(a)
     param_relline(6) = real(theta0grad)
     param_relline(7) = real(rmin)
     param_relline(8) = real(rmax)
     param_relline(9) = 0.e0  ! z
     param_relline(10) = limb_law  
     
  else if (model_type.eq.2) then

     ! READ Parameter Values
     h=param(1)
     a=param(2)
     theta0grad=param(3)
     rmin=param(4)
     rmax=param(5)
     limb_law=int(param(6))
     gamma=param(7)
     
     
     ! WRITE Parameter values (for RELBASE routine)
     param_relline(2) = real(h)
     param_relline(3) = real(a)
     param_relline(4) = real(theta0grad)
     param_relline(5) = real(rmin)
     param_relline(6) = real(rmax)
     param_relline(7) = 0.e0  ! z
     param_relline(8) = limb_law  
     param_relline(9) = real(gamma)

  else if (model_type.eq.3) then

     ! READ Parameter Values
     a=param(1)
     theta0grad=param(2)
     rmin=param(3)
     rmax=param(4)
     limb_law=int(param(5))
     gamma=param(6)
     hbase=param(7)
     htop=param(8)
     vbase=param(9)
     v100=param(10)
     
     
     ! WRITE Parameter values (for RELBASE routine)
     param_relline(2) = real(a)
     param_relline(3) = real(theta0grad)
     param_relline(4) = real(rmin)
     param_relline(5) = real(rmax)
     param_relline(6) = 0.e0
     param_relline(7) = limb_law  
     param_relline(8) = real(gamma)
     param_relline(9) = real(hbase)
     param_relline(10) = real(htop)
     param_relline(11) = real(vbase)
     param_relline(12) = real(v100)

  end if


END SUBROUTINE INIT_PAR_CONV !  ++++++++++++++++++++++++++++++++++++++++



!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
subroutine check_caching(param,num_param,photar,ear,ne,is_cached,same_input)
!  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  ! this routine checks if the parameters from a previous evalution 
  ! changed and tries to use previously calculated values for the current
  ! evalution to save time. These cached values are also administrated 
  ! and initialzied in this routine.

  !---------------------------------
  use par
  use par_conv
  implicit none

  integer :: num_param,i,ne
  real :: param(num_param),photar(ne),ear(0:ne)
  logical :: is_cached, same_input
  logical :: cache_relconv = .true.
  !---------------------------------

  ! == 1 == check if the same model was evaluated before or has to be initialized

  if (model_type.eq.cached_model_type ) then
     ! model was already evalated before
     cache_relconv = .false.
  else
     cache_relconv = .true.
     ! now we will cache the model, so lets also store this
     ! information for the next iteration
     cached_model_type = model_type
  endif


  ! ==== 2 ==== although model has been evaluted before, we need to
  !             check if anything changed
  if (.not.cache_relconv) then
     ! == 2a == check if any parameters changed
     ! at this point we can also check if the model evaluated before
     ! was the concolution model or the line model (line models have
     ! one more parameter)
     
     if (size(cached_params).eq.num_param) then
        do i=1,num_param
           if (cached_params(i).ne.param(i)) then
              cache_relconv = .true.
              exit
           end if
        enddo
     else
        cache_relconv = .true.
     end if

     ! == 2b == check if the energy grid changed 
     !          (not only number of elements, but also the values !!)
     if (size(cached_ear).eq.(ne+1)) then
        do i=0,ne
           if (cached_ear(i).ne.ear(i)) then
              cache_relconv = .true.
              exit
           endif
        end do
     else
        cache_relconv = .true.
     endif

  endif
 

  ! ==== 3 ==== Now we are able to decide if relline is alread cached or
  !         if no. If it's not cached, we have to do the caching, now!
  if (cache_relconv) then ! == 3a == relline IS NOT CACHED
     
     is_cached = .false.

     ! == 3b == dealloacte all arrays which exist
     if (allocated(cphotar)) then
        deallocate (cphotar)
     endif
     if (allocated(cached_params)) then
        deallocate (cached_params)
     endif
     if (allocated(cached_ear)) then
        deallocate (cached_ear)
     endif
     if (allocated(cached_input)) then
        deallocate (cached_input)
     endif
        

     ! == 3c == now allocate them with the correct dimension
     ! allocate the arrays for the cached values
     allocate (cached_params(num_param))
     allocate (cphotar(ne))
     allocate (cached_input(ne))
     allocate (cached_ear(0:ne))
     cached_ne = ne


     ! == 4d == finally cache the parameter and ear values
     do i=1,num_param
        cached_params(i) = param(i)
     enddo
     do i=0,ne
        cached_ear(i) = ear(i)
     enddo
     do i=1,ne
        cached_input(i) = photar(i)
     enddo



  else ! == 4 == relline IS CACHED =================================== !  
     
     is_cached = .true.

     ! did the input flux change?
     same_input = .true.
     do i=1,ne
        if (cached_input(i).ne.photar(i)) then
           same_input = .false.
           exit
        end if
     enddo

     if (.not.same_input) then
        ! input flux is not cached -> cache it for the next iteration
        do i=1,ne
           cached_input(i) = photar(i)
        enddo
     endif
     
  end if
  
  return

  
end subroutine check_caching !  ++++++++++++++++++++++++++++++++++
